from torii.data import Struct
from torii.exception import ToriiException
from torii.data.torii_object import ToriiObject


class Application(ToriiObject):

    _DICT_ENTRIES = ['phaseScripts', 'clusterScripts']

    def __init__(self, json, service, details=False):
        ToriiObject.__init__(self, json, service)

        self._details = details


    def has_details(self):
        return self._details

    def __assert_configurable(self):
        if not self._details:
            raise ToriiException('Cannot configure application without details')

    def __str__(self):
        """ String converter """
        if hasattr(self, 'name') and self.name and hasattr(self, 'version') and self.version:
            return '{2} {0}:{1}'.format(self.name, self.version, type(self))
        else :
            return '{1} {0}'.format(self.id, type(self))

    def clear_clusters(self):
        """
        Clears the cluster configuration
        :return:
        """
        self.__assert_configurable()
        self.clusters = []
        self.clusterScripts = {}

    def set_cluster(self, cluster, environment=None):
        """
        Tells that the application cannot run on a given cluster
        :param cluster: the cluster to unset
        :param environment: the configuration script
        :return:
        """
        self.__assert_configurable()

        self.clusters = [ref for ref in self.clusters if ref.id != cluster.id]
        self.clusters.append(cluster.ref)

        if environment is not None:
            self.clusterScripts[cluster.id] = environment
        else:
            self.clusterScripts.pop(cluster.id, None)


    def unset_cluster(self, cluster):
        """
        Tells that the application cannot run on a given cluster
        :param cluster: the cluster to unset
        :return:
        """
        self.__assert_configurable()

        if isinstance(cluster, ToriiObject):
            id = cluster.id
        else:
            id = cluster

        self.clusters = [ref for ref in self.clusters if ref.id != id]


    def set_phase(self, name, script=None, cardinality=None, **kwargs):
        """
        Configure a  phase

        :param name:
        :param script:
        :param cardinality:
        :param kwargs:
        :return:
        """
        self.__assert_configurable()

        phase = next(phase for phase in self.phases if phase.name == name)
        if not phase:
            phase = Struct({'name': name})
            self.phases.append(phase)

        # activate the phase if its cardinality has been
        if cardinality is None and hasattr(phase, 'cardinality') and phase.cardinality == 0:
            phase.cardinality = 1

        if kwargs:
            for key, value in kwargs.items():
                setattr(phase, key, value)


        if script is not None:
            self.phaseScripts[name] = script
        else:
            self.phaseScripts.pop(name, None)


    def unset_phase(self, name):
        """
        Make sure the phase is not executed when the application is run
        :param name:
        :return:
        """
        self.__assert_configurable()

        phase = next(phase for phase in self.phases if phase.name == name)
        if phase:
            setattr(phase, 'cardinality', 0)



from datetime import datetime as dt
import copy
import sys

from bson import ObjectId

from torii.data.struct import Struct, json_equals, json_diff
from torii.exception import ToriiException


class ToriiObject(Struct):
    """
    Wrapper for the torii objects
    """

    _DEEP_COPY_ORIG = True

    def __init__(self, json={}, service=None):

        if isinstance(json, ToriiObject):
            Struct.__init__(self, json.to_json())
            # always change the Id on copy
            self._id = str(ObjectId())
            self._orig = json._orig

        else:
            Struct.__init__(self, json)
            if self.__class__._DEEP_COPY_ORIG:
                self._orig = copy.deepcopy(json)
            else:
                self._orig = None

        # init the id if it was not
        if not hasattr(self, '_id'):
            self._id = str(ObjectId())

        self._service = service

    def from_json(self, json, transient_entries=None, wipe=True):
        try:
            Struct.from_json(self, json, transient_entries=transient_entries, wipe=wipe)
            if self.__class__._DEEP_COPY_ORIG:
                self._orig = copy.deepcopy(json)
            else:
                self._orig = None
        except:
            raise ToriiException('Failed convert ToriiObject \'{0}\' to Json'
                                 .format(json['name'] if 'name' in json else 'unknown')) \
                .with_traceback(sys.exc_info()[2])

    def to_json(self, ref=False, transient_entries=None):
        try:
            if ref:
                return {'id': self.id, 'name': self.name}
            else:
                json = super(ToriiObject, self).to_json(transient_entries=transient_entries)
                if 'id' in json:
                    del json['id']
                json['_id'] = self.id
                return json
        except:
            raise ToriiException('Failed convert {0} from Json'.format(self)) \
                .with_traceback(sys.exc_info()[2])

    def has_changed(self):
        return not hasattr(self, '_orig') or not json_equals(self._orig, self.to_json())

    @property
    def id(self):
        return self._id

    @property
    def diff(self):
        if not hasattr(self, '_orig'):
            return self

        diff = json_diff(self._orig, self.to_json())

        # add the id in the diff only if not empty
        if diff:
            diff['id'] = self.id

        return diff

    def clear_id(self):
        self.id = str(ObjectId())

    @property
    def orig(self):
        return self._orig

    @property
    def ref(self):
        return Struct({'id': self.id, 'name': self.name})

    def match(self, obj):
        """
        Does the object ID match another object or reference or ID string
        :param obj: the object to match
        :return: boolean
        """
        if isinstance(obj, Struct):
            return self.id == obj.id
        else:
            return self.id == obj

    def create(self):
        return self._service.create(self)

    def update(self, **kwargs):
        if (self.has_changed()):
            return self._service.update(self, **kwargs)
        else:
            return self._service.refresh(self)

    def patch(self, *args):
        return self._service.patch(self.id, *args)

    def refresh(self):
        return self._service.refresh(self)

    def action(self, action=None, params=None, update=True, send_payload=True, raw_result=True):
        return self._service.action(action=action,
                                    obj=self,
                                    params=params,
                                    update=update,
                                    send_payload=send_payload,
                                    raw_result=raw_result)

    def snapshot(self):
        return self._service.action('snapshot', self)

    def revert(self):
        return self._service.action('revert', self)



import logging
from operator import itemgetter

from bson import ObjectId

import torii as Torii
from torii.dao.abstract_dao import AbstractDao


class UserManager(AbstractDao):

    COLLECTION_USERS = 'users'
    COLLECTION_SESSION = 'sessions'
    COLLECTION_TEAMS = 'teams'

    CACHED_TEAMS = []

    def __init__(self, torii):
        self.torii = torii
        self.torii_db = torii.mongo_database_torii
        self._logger = logging.getLogger('torii')
        
        self.user = None

    def check_user_permission(self, bo_class_name, read=True, write=False) -> tuple[bool, str]:
        """
        Checks User permissions for the collection in context.

        Checks if class is readable/writeable and if User is in the same team as the said class,
        from Torii.bo_classes collection.

        :param bo_class_name: the nameofthe BO class to check.
        :param read: Set True if User is reading any document, otherwise False.
        :param write: Set True if User is writing, updating or deleting any document, otherwise False.
        :return: A bool to identify if User has permission, and a str explaining the reason.
        """
        if not self.user:
            self.update_user()

        if self.user.admin:
            return True, 'User is admin.'

        bo_class = next((item for item in self.CACHED_BO_CLASSES if item['name'] == bo_class_name), None)
        if not bo_class:
            collection_classes = self.torii_db.get_collection(self.COLLECTION_BO_CLASSES)
            bo_class = self.find_one(collection_classes, {'name': bo_class_name})
            if bo_class is not None:
                self.CACHED_BO_CLASSES.append(bo_class)

        if read and not bo_class['fullReadable']:
            return False, 'Class is not readable.'

        if write and not bo_class['fullWritable']:
            return False, 'Class is not writeable.'

        team = self.check_team_access(bo_class)

        if team is None:
            return False, 'User does not have team permission.'

        return True, 'User checks out.'

    def check_team_access(self, document: dict):
        """
        Helper method.

        Check if User is in the same team as the document in question.

        :param document: JSON-like object representing the document.
        :return: The given document if teams match, otherwise None.
        """
        if not document:
            return None
        for team in self.user.teams:
            # Check if team ID match
            if team['id'] in map(itemgetter('id'), document['teams']):
                team_id = ObjectId(team['id'])
                # Check if team is enabled
                matched_team = next((item for item in self.CACHED_TEAMS if item['_id'] == team_id), None)
                if not matched_team:
                    collection = self.torii_db.get_collection(self.COLLECTION_TEAMS)
                    matched_team = self.find_one(collection, {'_id': team_id})
                    self.CACHED_TEAMS.append(matched_team)
                if matched_team['status'] == 'ENABLED':
                    return document
        return None

    def remove_no_access_documents(self, documents: list) -> list:
        """
        Helper method.

        Removes all documents in the list that the User has no access to.

        :param documents: List of JSON-like objects that represent documents.
        :return: The filtered list.
        """
        if not documents:
            return []
        return [document for document in documents if self.check_team_access(document)]

    def update_user(self) -> None:
        """
        Private helper method.

        Get the User by ID from the current Torii session.

        :return: The User as a JSON-like object.
        """
        user_id = ObjectId(self.torii.session.user['id'])
        collection = self.torii_db.get_collection(self.COLLECTION_USERS)
        user = self.find_one(collection, {'_id': user_id})
        user['_id'] = str(user['_id'])
        self.user = Torii.ToriiObject(user)



import ast
import copy
import re
import sys

import numpy as np
from bson import ObjectId

from torii.exception import ToriiException


def camel2snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def snake2camel(name):
    components = name.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])


def cast_value(value):
    """
    Try to cast to float, integer, dict, bool if possible
    ast.literal_eval(node_or_string)
    Safely evaluate an expression node or a string containing a Python
    expression. The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None.

    :param value: the value to cast
    :return:
    """

    # for a simple string...
    if isinstance(value, str):
        # ast gives an best effort evaluator
        try:
            intermediate = ast.literal_eval(value)
        except:
            intermediate = value

        # ...but it is not recursive enough and it does not try to evaluate the leafs
        # it as detected so...
        if isinstance(intermediate, dict) or isinstance(intermediate, list):
            # ... if it has returned an structure, we reapply our cast_value recursively
            return cast_value(intermediate)
        else:
            # ... in case of a simple type, we stop the recursion
            return intermediate

    # for dictionaries...
    elif isinstance(value, dict):
        casted_value = {}
        for key, subvalue in value.items():

            # ... we apply the cast rescursively
            try:
                casted_subvalue = cast_value(subvalue)
            except:
                # ... and we best effort in case of pb
                casted_subvalue = subvalue

            casted_value[key] = casted_subvalue

        return casted_value

    # for the lists we use the same policy
    elif isinstance(value, list):
        casted_value = []

        for subvalue in value:
            try:
                casted_subvalue = cast_value(subvalue)
            except:
                casted_subvalue = subvalue

            casted_value.append(casted_subvalue)
        return casted_value

    # for anything else we consider there is nothing more to cast
    else:
        # ... so we end the recursion
        return value


def json_equals(old, new):
    if not isinstance(old, type(new)):
        return old == new
    elif isinstance(old, dict):
        if set(old.keys()) != set(new.keys()):
            return False

        for key, value in new.items():
            old_value = old[key]
            if not json_equals(old_value, value):
                return False
        return True
    elif isinstance(old, list):
        if len(old) != len(new):
            return False
        i = 0
        for value in new:
            old_value = old[i]
            if not json_equals(old_value, value):
                return False
            i += 1

        return True
    else:
        return old == new


def json_diff(old, new):
    diff = {}
    if not isinstance(old, type(new)):
        diff = new
    elif isinstance(old, dict):
        for key, value in new.items():
            if key not in old or not json_equals(old[key], value):
                diff[key] = value

        for key in old.keys():
            if key not in new:
                diff[key] = None

    elif isinstance(old, list):
        if not json_equals(old, new):
            diff = new

    return diff


def json_to_struct(obj):
    if isinstance(obj, dict):
        return Struct(obj)
    elif isinstance(obj, list):
        if len(obj) > 10 and not isinstance(obj[0], dict):
            return obj
        else:
            return [json_to_struct(item) for item in obj]
    elif isinstance(obj, str):
        return obj
    else:
        return copy.deepcopy(obj)


def struct_to_json(obj):
    if isinstance(obj, Struct):
        return obj.to_json(ref=True)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, list):
        if len(obj) > 10 and not isinstance(obj[0], dict) and not isinstance(obj[0], Struct):
            return obj
        else:
            return [struct_to_json(item) for item in obj]
    elif isinstance(obj, str):
        return obj
    else:
        return copy.deepcopy(obj)


class Struct(object):
    # list of entries that must remain as dictionary
    _DICT_ENTRIES = []

    # list of entries that must not appear in json
    _TRANSIENT_ENTRIES = []

    def __init__(self, json={}):
        """
        Create a proper hierarchical structure base on json-like dictionary
        """

        if json is not None:
            if '_id' in json and json['_id'] is not None:
                if isinstance(json['_id'], ObjectId):
                    json['_id'] = str(json['_id'])
                if '$oid' in json['_id']:
                    setattr(self, '_id', json['_id']['$oid'])
                else:
                    setattr(self, '_id', json['_id'])

            self.from_json(json)

    def __getattr__(self, key):
        """
        Override 'getattr()' and dot operator
        :param key:  the attribute key
        :return:
        """

        if key == 'id' and '_id' in self.__dict__:
            # we differentiate the '_id' of a ToriiObject ('_id' in json)
            # from the id that can be found in the sub documents ('id' in json)
            return self.__dict__['_id']
        elif key in self.__dict__:
            return self.__dict__[key]

        raise AttributeError

    def __setattr__(self, key, value):
        """
        Override 'setattr()' and dot operator
        :param key:  the attribute key
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __delattr__(self, key):
        """
        Override 'delattr()'
        :param key:  the attribute key
        :return:
        """
        if key in self.__dict__:
            del self.__dict__[key]

    def __iter__(self):
        """
        iterate over the attributes
        :return: an iterator
        """
        return iter(self.keys())

    def __getitem__(self, key):
        """
        Override [] operator
        :param key: the attribute key
        :return:
        """
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        """
        Override [] operator assignment
        :param key:
        :param value:
        :return:
        """
        self.__setattr__(key, value)

    def __delitem__(self, key):
        """
        Override [] operator assignment
        :param key:
        :return:
        """
        self.__delattr__(key)

    def __contains__(self, key):
        """
        Override operator 'in'
        :param key: the attribute key
        :return:
        """
        try:
            self.__getattr__(key)
        except:
            return False
        return True

    def __len__(self):
        """
        Override 'len()'
        :return:
        """
        return len(self.keys())

    def __str__(self):
        """
        String converter
        """
        if hasattr(self, 'name') and self.name:
            return '{1} {0}'.format(self.name, type(self))
        elif hasattr(self, 'num') and self.num:
            return '{1} {0}'.format(self.num, type(self))
        elif hasattr(self, 'id') and self.id:
            return '{1} {0}'.format(self.id, type(self))
        else:
            return '{0}'.format(type(self))

    def __eq__(self, other):
        """
        Override '==' operator
        :param other:
        :return:
        """
        if isinstance(other, self.__class__):
            return json_equals(self.to_json(), other.to_json())
        elif isinstance(other, dict):
            return json_equals(self.to_json(), other)
        else:
            return False

    def __ne__(self, other):
        """
        Override '!=' operator
        :param other:
        :return:
        """
        return not self.__eq__(other)

    def keys(self):
        """
        Override 'keys()'
        """

        return [key for key in self.__dict__.keys() if isinstance(key, int) or not key.startswith('_')]

    def items(self):
        """
        Override 'items()'
        """
        return [(key, self.__getattr__(key)) for key in self.keys()]

    def from_json(self, json, transient_entries=None, wipe=True):
        """
        Update the structure based on a Json compatible dictionary
        """
        key = None
        try:
            if wipe:
                # remove the former attributes
                for key in Struct.keys(self):
                    if key.startswith('_'):
                        pass
                    elif key in self.__class__._TRANSIENT_ENTRIES:
                        pass
                    elif transient_entries is not None and key in transient_entries:
                        pass
                    else:
                        self.__delattr__(key)

            # re-init them
            for key, value in json.items():
                if isinstance(key, str) and key.startswith('_'):
                    pass
                elif key in self.__class__._TRANSIENT_ENTRIES:
                    pass
                elif transient_entries is not None and key in transient_entries:
                    pass
                elif key not in self._DICT_ENTRIES:
                    Struct.__setattr__(self, key, json_to_struct(value))
                else:
                    Struct.__setattr__(self, key, value)
        except:
            raise ToriiException('Failed convert struct from Json on key \'{0}\''
                                 .format(key)).with_traceback(sys.exc_info()[2])

    def to_json(self, ref=False, transient_entries=None):
        """
        Export the structure as a Json compatible dictionary
        :param ref: export as reference (only for ToriiObject)
        """
        json = {}

        key = None
        try:
            if hasattr(self, '_id') and (not transient_entries or '_id' not in transient_entries):
                json['_id'] = self.id

            for key, value in self.items():
                if key in self._TRANSIENT_ENTRIES:
                    pass
                elif transient_entries is not None and key in transient_entries:
                    pass
                else:
                    json[key] = struct_to_json(value)

        except:
            raise ToriiException('Failed convert struct to Json on key \'{0}\''
                                 .format(key)).with_traceback(sys.exc_info()[2])

        return json




from torii.data import *
from torii.services.torii_service import ToriiService


class BusinessObjectService(ToriiService):

    def __init__(self, torii, bo_name, object_class=BusinessObject):
        """
        Create the BO service

        :param torii:
        :param bo_name:
        """
        ToriiService.__init__(self, torii=torii, name='bo/{0}'.format(bo_name), object_class=object_class)



import logging
import re
import sys
import json as json_lib

import requests
from requests_toolbelt import MultipartEncoder

from torii.exception import ToriiException


class Service(object):

    def __init__(self, torii, name, base_path=None):
        """
        Create a generic Gateway service
        """

        self.name = name
        self.base_path = base_path if base_path else name
        self._torii = torii
        self.dao_manager = torii.dao_manager
        self._logger = logging.getLogger('torii')

    def __str__(self):
        """
        Convert to string
        """
        if self.name is self.base_path:
            return 'service {0}'.format(self.name)
        else:
            return 'service {0} ({1})'.format(self.name, self.base_path)

    def get_url(self, path):
        """
        Get the Rest command URL corresponding to the relative path
        :param path: the relative path
        :return: the URL
        """
        if path.startswith('/'):
            return self._torii.properties['base_url'] + path
        else:
            return self._torii.properties['base_url'] + '/rest/' + self.base_path + '/' + path

    def get_simple_url(self, path):
        """
        Get the API path (service path + method path)

        :param path: the relative path
        :return: the API path
        """
        if path.startswith('/'):
            return path
        else:
            return self.base_path + '/' + path

    def request(self, method, path='', params=None, json=None, data=None, 
                raw_result = False, files=None, anonymous=False, timeout=None):
        """
        Submit a standard REST request
        :param method: GET, PUT, POST or DELETE
        :param path: the relative path
        :param params: the request parameters
        :param json: the Json payload
        :param data: the raw payload
        :param raw_result: return the raw result
        :param anonymous: send the request outside the HTTP session

        :return: the result as Json
        """
        try:

            headers = {}
            if data and isinstance(data, MultipartEncoder):
                headers['Content-Type'] = data.content_type

            if anonymous:
                r = requests.request(method=method,
                                     url=self.get_url(path),
                                     json=json,
                                     data=data,
                                     files=files,
                                     params=self._prepare_params(params),
                                     headers=headers,
                                     timeout=timeout)
            else:
                r = self._torii.http_session.request(method=method,
                                                     url=self.get_url(path),
                                                     json=json,
                                                     data=data,
                                                     files=files,
                                                     params=self._prepare_params(params),
                                                     headers=headers,
                                                     timeout=timeout)

            if r.status_code == 200:
                if method in ['GET', 'POST', 'PUT', 'PATCH'] and not raw_result:
                    return r.json()
                else:
                    return str(r.content)
            elif r.status_code in [204]:
                return None
            else:
                message = 'Failed to request {0} {1} on {2} (code = {3})' \
                          '\n\tserver message: \n\t\t{4}'\
                    .format(method, path, str(self), r.status_code, r.reason)\
                    .replace('\0', '')

                content = str(r.content).replace('\0', '')
                if '{' in content and '}' in content:
                    try:
                        raw_json_str = content[content.find('{'):content.rfind('}') + 1]
                        raw_json_str = raw_json_str.replace('\\n    "', '"').replace('\\n}', '}')
                        content = json_lib.loads(raw_json_str)
                        if 'message' in content:
                            message += '\n\tJava error: ' + content['message']
                        if 'detail' in content:
                            stacktrace_str = content['detail'].replace('\\n', '\n\t').replace('\\t', '\t')
                            message += '\n\tJava stacktrace: \n\t\t{0}'.format(stacktrace_str)
                    except:
                        message += '\n\t\t{0}'.format(content)
                else:
                    message += '\n\t\t{0}'.format(content)

                raise ToriiException(message)

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to request {0} {1} on {2} \n{3}'
                                 .format(method, path, str(self), e)).with_traceback(sys.exc_info()[2])

    def _prepare_params(self, params):
        prepared = {}
        if params is not None:
            for name, value in params.items():
                if value is True:
                    value = 'true'
                elif value is False:
                    value = 'false'
                prepared[name] = value
        return prepared

    def request_get(self, path='', params=None, files=None, timeout=None):
        """
        Simplified GET Rest request
        """
        try:
            service = self.dao_manager.run_service('GET', self.get_simple_url(path), params=params, files=files, timeout=timeout)
        except ToriiException as e:
            raise e
        except Exception as e:
            service = None

        return service if service is not None else self.request('GET', path=path, params=params, files=files, timeout=timeout)

    def request_post(self, path='', params=None, json=None, data=None, files=None, timeout=None):
        """
        Simplified POST Rest request
        """
        try:
            service = self.dao_manager.run_service('POST', self.get_simple_url(path), params=params, json=json, data=data, files=files, timeout=timeout)
        except ToriiException as e:
            raise e
        except Exception as e:
            service = None

        return service if service is not None else self.request('POST', path=path, params=params, json=json, data=data, files=files, timeout=timeout)

    def request_put(self, path='', params=None, json=None, data=None, files=None, timeout=None):
        """
        Simplified PUT Rest request
        """
        try:
            service = self.dao_manager.run_service('PUT', self.get_simple_url(path), params=params, json=json, data=data, files=files, timeout=timeout)
        except ToriiException as e:
            raise e
        except Exception as e:
            service = None

        return service if service is not None else self.request('PUT', path=path, params=params, json=json, data=data, files=files, timeout=timeout)

    def request_delete(self, path='', params=None, timeout=None):
        """
        Simplified DELETE Rest request
        """
        try:
            service = self.dao_manager.run_service('DELETE', self.get_simple_url(path), params=params, timeout=timeout)
        except ToriiException as e:
            raise e
        except Exception as e:
            service = None

        if service is None: self.request('DELETE', path=path, params=params, timeout=timeout)

    def request_patch(self, path='', params=None, json=None, timeout=None):
        """
        Simplified PATCH Rest request
        """
        try:
            service = self.dao_manager.run_service('PATCH', self.get_simple_url(path), params=params, json=json, timeout=timeout)
        except ToriiException as e:
            raise e
        except Exception as e:
            service = None

        return service if service is not None else self.request('PATCH', path=path, params=params, json=json, timeout=timeout)



import base64
import glob
import json as jsonlib
import os
import re
import shutil

import sys
import tempfile
import zipfile

from torii.exception import ToriiException
from torii.data import *
from torii.data import json_to_struct
from torii.services.torii_service import ToriiService



def _underscore_to_camelcase(value):
    def camelcase():
        yield str.lower
        while True:
            yield str.capitalize

    c = camelcase()
    return "".join(next(c)(x) for x in value.split("_") if x)


def _read_script_content(path, dos2unix=True):
    with open(path, 'r') as file:
        content = file.read()

        if dos2unix:
            content = content.replace('\r\n', '\n')

        return content


def _insert_functions(script, functions):
    result = ''

    in_header = True
    for line in script.splitlines():

        # we look for the first line that is NOT in the header
        if in_header:
            strip_line = line.lstrip()
            header_line = len(strip_line) == 0 \
                          or strip_line.startswith('#') \
                          or strip_line.startswith('import') \
                          or strip_line.startswith('from')

            if not header_line:
                # ... at the end of the header, we insert the function definitions
                in_header = False
                result = result + functions + '\n'

        # we add the lines of the original script one by one
        result = result + line + '\n'

    return result


class ApplicationService(ToriiService):

    def __init__(self, torii):
        """
        Create the application service
        """
        ToriiService.__init__(self, torii=torii, name='applications', object_class=Application)


    def get(self, id=None, details=False, **kwargs):
        if id and details:
            json = self.request_get(path='{0}'.format(id), params={'details':'true'})
            return Application(json, self, details=True)
        elif details:
            application = ToriiService.get(self, **kwargs)
            return self.get(application.id, details=True) if application else None
        else:
            return ToriiService.get(self, id, **kwargs)


    def update(self, application, **kwargs):
        if isinstance(application, Application) and (application.has_details()
                                                     or 'details' in kwargs and kwargs['details']):
            if application.has_changed():
                # update the remote version
                kwargs['editMode'] = True
                json = self.request_post(path='{0}'.format(application.id), json=application.to_json(), params=kwargs)
            else:
                # update local version
                json = self.request_get(path='{0}'.format(application.id))

            if json:
                application.from_json(json)
        else:
            application = ToriiService.update(self, application, editMode=False)

        return application

    def __export_application(self, application, path=None, zip=False):
        try:
            if not application.has_details():
                raise ToriiException('Cannot export {0}, no details'.format(application))

            if zip:
                dir = tempfile.mkdtemp()
            else:
                if not os.path.exists(path):
                    os.makedirs(path)
                dir = path

            # write the application JSON
            json = application.to_json()
            # ... the verbose fields are removed from the json to be written independently
            for field in ['icon', 'clusters', 'clusterScripts', 'phaseScripts', 'validationScript']:
                json.pop(field, None)
            with open(os.path.join(dir, 'definition.json'), 'w+') as outfile:
                jsonlib.dump(json, outfile, indent=4, sort_keys=True)

            # decode and write the application icon
            if hasattr(application, 'icon'):
                try:
                    regex = re.compile('data:image/(?P<type>[a-z]+);base64,(?P<content>.*)')
                    match = regex.match(application.icon)
                    if match:
                        icon_path = os.path.join(dir, 'icon.' + match.groupdict()['type'])
                        with open(icon_path, 'w+b') as outfile:
                            content = base64.b64decode(match.groupdict()['content'])
                            outfile.write(content)
                except:
                    self._logger.warn('Failed to extract icon for {}'.format(application))

            # write the cluster environment scripts
            clusters_dir = os.path.join(dir, 'clusters')
            os.makedirs(clusters_dir)
            for id, script in application.clusterScripts.items():
                try:
                    cluster = self._torii.clusters.get(id)
                    if cluster:
                        with open(os.path.join(clusters_dir, cluster.name), 'w+') as outfile:
                            outfile.write(script)
                except:
                    self._logger.warn('Failed to write cluster environment for {}'.format(application))

            # write the validation scripts
            scripts_dir = os.path.join(dir, 'scripts')
            os.makedirs(scripts_dir)
            if hasattr(application, 'validationScript') and application.validationScript:
                with open(os.path.join(scripts_dir, 'validation.py'), 'w+') as outfile:
                    outfile.write(application.validationScript)

            if hasattr(application, 'interfaceScript') and application.interfaceScript:
                with open(os.path.join(scripts_dir, 'interface.js'), 'w+') as outfile:
                    outfile.write(application.interfaceScript)

            # write the phases scripts
            phases_dir = os.path.join(dir, 'phases')
            os.makedirs(phases_dir)
            for name, script in application.phaseScripts.items():
                if name is 'monitor':
                    # currently the monitor script is still in the phases ... but it should not
                    filepath = os.path.join(scripts_dir, name)
                else:
                    filepath = os.path.join(phases_dir, name)

                with open(filepath, 'w+') as outfile:
                    outfile.write(script)

            if zip:
                # go to the temp dir
                cwd = os.getcwd()
                os.chdir(dir)

                # zip its content
                ziph = zipfile.ZipFile(path, 'w', zipfile.ZIP_DEFLATED)
                for root, dirs, files in os.walk('.'):
                    for file in files:
                        ziph.write(os.path.join(root, file))

                # go back to the current dir
                os.chdir(cwd)

                # and cleanup
                shutil.rmtree(dir)


        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to export {0}\n\t{1}'.format(application, e)).with_traceback(sys.exc_info()[2])


    def __import_application(self, path=None, dos2unix=True):
        try:
            if not os.path.exists(path):
                raise ToriiException('Cannot import application from \'{0}\', no such file or directory'
                                     .format(path))

            if os.path.isdir(path):
                root_dir = path
            else:
                root_dir = tempfile.mkdtemp()

                try:
                    ziph = zipfile.ZipFile(path, 'r')
                    ziph.extractall(root_dir)
                    ziph.close()
                except Exception as e:
                    raise ToriiException('Failed to import application from \'{0}\'\n\t{1}'.format(path, e)).with_traceback(sys.exc_info()[2])

            # the application description can be in a subdirectory
            dir = None
            for root, dirs, files in os.walk(root_dir):
                if 'application.json' in files or 'definition.json' in files:
                    dir = root

            if dir is None:
                raise ToriiException('Cannot find application.json in {0}'.format(path))


            application = self.__import_application_json(dir)

            # consolidate the refrences appearing in the application
            self.__consolidate_application_references(application)

            # configure the clusters environment
            self.__import_application_environement(application, dir, dos2unix)

            # import the phase scripts
            self.__import_application_phases(application, dir, dos2unix)

            # import the other scripts (validation and monitoring)
            self.__import_application_scripts(application, dir, dos2unix)

            # import each module from the modules directory
            modules_dir = os.path.join(dir, 'modules')
            if os.path.isdir(modules_dir):
                for module_name in os.listdir(modules_dir):
                    self.__import_application_module(application,
                                                     module_name,
                                                     os.path.join(modules_dir, module_name),
                                                     dos2unix)

            # get the application icon
            icon_paths = glob.glob(os.path.join(dir, 'icon.*'))
            if icon_paths and os.path.isfile(icon_paths[0]):
                image_type = icon_paths[0].split('.')[-1].lower()
                with open(icon_paths[0], 'rb') as file:
                    content = file.read()
                    if content:
                        application.icon = 'data:image/{0};base64,{1}'\
                            .format(image_type, str(base64.b64encode(content), 'utf8'))

            ### legacy fixes ###
            # create a default phase in case of incoherence
            if 'monitor' in application.phaseScripts \
                    and not any(phase.name == 'monitor' for phase in application.phases):
                application.phases.append(Struct({'name': 'monitor'}))

            # remove the 'valid' attribute from the inputs
            for input in application.inputs:
                if hasattr(input, 'valid'):
                    delattr(input, 'valid')

            return application

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to import application from \'{0}\'\n\t{1}'.format(path, e)).with_traceback(sys.exc_info()[2])


    def __import_application_json(self, dir):
        # read the application JSON
        if  'application.json' in os.listdir(dir):
            json_path = os.path.join(dir, 'application.json')
        elif 'definition.json' in os.listdir(dir) :
            json_path = os.path.join(dir, 'definition.json')
        else:
            raise ToriiException('Cannot import application from, missing definition file in {0}'
                                 .format(dir))

        with open(json_path) as file:
            json = jsonlib.load(file)

        application = Application(json, self, True)

        # get complementary definitions from other json files (inputs.json, )
        extra_json_path = {name.replace('.json', '', 1): os.path.join(dir, name)
                         for name in os.listdir(dir)
                           if name.endswith('.json') and name not in ['application.json', 'definition.json']}
        for name, path in extra_json_path.items():
            # put the compementory description in the corresponding field
            with open(path) as file:
                json = jsonlib.load(file)
                struct = json_to_struct(json)
                field = _underscore_to_camelcase(name)

                setattr(application, field, struct)


        # default
        if not hasattr(application, 'inputs'):
            setattr(application, 'inputs', [])

        return application


    def __import_application_module(self, application, module_name, module_dir, dos2unix):

        module_prefix = module_name + '__'

        # get the inputs from the module
        inputs_path = os.path.join(module_dir, 'inputs.json')
        if os.path.isfile(inputs_path):
            with open(inputs_path) as file:
                json = jsonlib.load(file)
                inputs = json_to_struct(json)

                for input in inputs:
                    if not input.name.startswith(module_prefix):
                        raise ToriiException('Invalid input name \'{0}\' in module \'{1}\' (must start with {2})'
                                             .format(input.name, module_name, module_prefix))

                application.inputs.extend(inputs)

        # append the validation script from the module
        validate_path = os.path.join(module_dir, 'validation.py')
        if os.path.isfile(validate_path):
            validate_functions = _read_script_content(validate_path, dos2unix)
            application.validationScript = _insert_functions(application.validationScript, validate_functions)

        # read the custom layout panels from the module
        panels_path = os.path.join(module_dir, 'custom_layout.json')
        if os.path.isfile(panels_path):
            with open(panels_path) as file:
                json = jsonlib.load(file)
                module_panels = json_to_struct(json)

                # make a map out of the panel or list of panels read from the json
                panel_map = {}
                if isinstance(module_panels, list):
                    for panel in module_panels:
                        panel_map[panel.panelId] = panel
                else:
                    panel_map[module_panels.panelId] = module_panels

                # check that the panel ids correspond to the module
                for panel in panel_map.values():
                    if not panel.panelId.startswith(module_prefix):
                        raise ToriiException('Invalid panel id \'{0}\' in module \'{1}\' (must start with {2})'
                                             .format(panel.panelId, module_name, module_prefix))

                # explore the panels and replace the module panel reference
                def replace_panel_references(panel, list, index):
                    if hasattr(panel, 'panelId') and panel.panelId in panel_map:
                        list[index] = panel_map[panel.panelId]
                    elif hasattr(panel, 'panels'):
                        for i in range(len(panel.panels)):
                            replace_panel_references(panel.panels[i], panel.panels, i)

                replace_panel_references(application.customLayout, None, 0)


    def __import_application_phases(self, application, dir, dos2unix):
        # list the phases scripts
        phases_dir = os.path.join(dir, 'phases')
        if os.path.isdir(phases_dir):
            # get all the files from the 'phases' dir
            phase_files = {os.path.splitext(filename)[0]: os.path.join(phases_dir, filename) for filename in os.listdir(phases_dir)}
        else:
            # or get all the 'app_<phase>' files from the root dir (legacy)
            phase_files = {name.replace('app_', '', 1): os.path.join(dir, name)
                         for name in os.listdir(dir)
                           if name.startswith('app_') and name not in ['validate', 'monitor']}

        # ... and associate the scripts with the phases
        phases = {}
        phase_list = []
        for phase in application.phases:
            if isinstance(phase, Struct):
                phases[phase.name] = phase
                phase_list.append(phase)
            elif isinstance(phase, str):
                # for legacy purpose we reconstitute a phase from a simple name
                #  - only the 'execute' phase uses User Defined Scheduler Options (UDSO)
                #  - we create a simple phase dependency
                phase_obj = Struct({'name': phase,
                                    'udso': (phase == 'execute'),
                                    'dependencies': [phase.name for phase in phase_list[-1:]]})
                phases[phase] = phase_obj
                phase_list.append(phase_obj)
        application.phases = phase_list

        # we don't overwrite 'phaseScripts' directly in case the script were in the Json
        if not hasattr(application, 'phaseScripts') or application.phaseScripts is None:
            application.phaseScripts = {}

        for name, path in phase_files.items():
            application.phaseScripts[name] = _read_script_content(path, dos2unix)

        return phases


    def __import_application_scripts(self, application, dir, dos2unix):
        # read the validation script
        scripts_dir = os.path.join(dir, 'scripts')
        if os.path.isfile(os.path.join(scripts_dir, 'validation.py')):
            application.validationScript = _read_script_content(os.path.join(scripts_dir, 'validation.py'), dos2unix)
        elif 'app_validate' in os.listdir(dir):
            # (legacy case)
            application.validationScript = _read_script_content(os.path.join(dir, 'app_validate'), dos2unix)

        # read the interface script
        if os.path.isfile(os.path.join(scripts_dir, 'interface.js')):
            application.interfaceScript = _read_script_content(os.path.join(scripts_dir, 'interface.js'), dos2unix)

        # and the monitoring script
        if os.path.isfile(os.path.join(scripts_dir, 'monitor')):
            application.phaseScripts['monitor'] = _read_script_content(os.path.join(scripts_dir, 'monitor'), dos2unix)
        elif 'app_monitor' in os.listdir(dir):
            application.phaseScripts['monitor'] = _read_script_content(os.path.join(dir, 'app_monitor'), dos2unix)



    def __import_application_environement(self, application, dir, dos2unix):
        # read the cluster environment scripts
        application.clear_clusters()
        clusters_dir = os.path.join(dir, 'clusters')
        cluster_scripts = {}
        if os.path.isdir(clusters_dir):
            for name in os.listdir(clusters_dir):
                script = _read_script_content(os.path.join(clusters_dir, name), dos2unix)
                cluster_scripts[re.sub('\.env$', '', name)] = script

        # ... and associate the scripts to the local clusters
        application.clusterScripts = {}
        clusters = self._torii.clusters.list()
        for cluster in clusters:
            script = None
            if cluster.id in cluster_scripts:
                script = cluster_scripts[cluster.id]
            elif cluster.name in cluster_scripts:
                script = cluster_scripts[cluster.name]
            elif 'default' in cluster_scripts:
                script = cluster_scripts['default']

            if script is not None:
                application.clusterScripts[cluster.id] = script
            application.clusters.append(cluster.ref)


    def __consolidate_application_references(self, application):
        # update references in users, teams and projects
        for ref_type in ['users', 'teams', 'projects']:
            if hasattr(application, ref_type):
                updated_refs = []

                # for each imported reference we try to find in torii a matching object
                for ref in getattr(application, ref_type):
                    object = None
                    try:
                        if ref.id:
                            object = getattr(self._torii, ref_type).get(ref.id)
                        elif ref.name:
                            object = getattr(self._torii, ref_type).get(name=ref.name)
                    except:
                        # ... if no matching object is found, we consider that the reference can be dropped
                        pass

                    if object is not None:
                        updated_refs.append(object.ref)

                setattr(application, ref_type, updated_refs)
            else:
                setattr(application, ref_type, [])


    def import_application(self, path=None, force=False, duplicate=False, dos2unix=True, **kwargs):
        """

        :param path:
        :return:
        """
        try:
            applications = []
            realPaths = glob.glob(path)

            for realPath in realPaths:
                application = self.__import_application(realPath, dos2unix)
                existing_application = None

                if duplicate:
                    application.clear_id()
                    application.status = 'DRAFT'
                else:
                    try:
                        existing_application = self.get(id=application.id)
                    except:
                        pass

                    if not force and existing_application is not None:
                        raise ToriiException('Cannot import \'{0}\', {1} already exist '.format(realPath, existing_application)
                                             + '(use \'force\' if you want to overwrite it)')

                # set the application fields explicitely specified by the user
                if kwargs:
                    for key, value in kwargs:
                        setattr(application, key, value)


                if existing_application:
                    # keep the authorizations
                    application.users    = existing_application.users
                    application.teams    = existing_application.teams
                    application.projects = existing_application.projects

                    # update the application
                    application.update(force=force, **kwargs)
                else:
                    # create the application
                    application.create(**kwargs)

                applications.append(application)

            if len(applications) == 0:
                raise ToriiException('Path {0} does not correspond to any application'.format(path))
            elif len(applications) == 1:
                return applications[0]
            else:
                return applications

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to import from \'{0}\'\n\t{1}'.format(path, e)).with_traceback(sys.exc_info()[2])

    def export_application(self, id=None, path=None, dir=None, zip=False, **kwargs):
        """
        Export one or more aplications
        :param id: the id of the single application to export
        :param path: the exact path to export to
        :param dir: the directory to export to (not compatible with path)
        :param zip: must we zip the application archive
        :param kwargs: other arguments to identify the applications to export
        :return:
        """

        try:
            # get the application ids to export
            if id is not None:
                ids = [id]
            else:
                ids = [app.id for app in self.list(**kwargs)]

            if len(ids) == 0:
                raise ToriiException('Cannot export, no application matching the parameters')

            if path and len(ids) > 1:
                raise ToriiException('Cannot export, many applications matching the parameters '
                                     'while a path has been specified')

            if not path and not dir:
                dir = os.getcwd()

            for id in ids:
                application = self.get(id=id, details=True)

                if path:
                    current_path = path
                else:
                    filename = '{0}-{1}.{2}'.format(application.name, application.version, application.id)\
                        .lower()\
                        .replace(' ', '_')\
                        .replace('/', '_')

                    if zip:
                        filename += '.zip'

                    current_path = os.path.join(dir, filename)

                self.__export_application(application, current_path, zip)

                return current_path

        except ToriiException:
            raise
        except Exception as e:
            raise ToriiException('Failed to export applications\n\t{1}'.format(application, e)).with_traceback(sys.exc_info()[2])




import inspect

import re

from torii.data import Struct
from torii.exception import ToriiException
from torii.data.torii_object import ToriiObject


def fun2script(function):

    imports = ''

    lines = inspect.getsourcelines(function)
    index = 0
    for line in lines[0]:
        if index == 0:
            # parse the function header
            m = re.search('(^ *)def (.*)\((.*)\) *:', line)
            indent = '^' + m.group(1)
            function_name = m.group(2)
            function_args = m.group(3)

            # rewrite the function header without indent
            function_def = 'def {0}({1}):\n'.format(function_name, function_args)
        else:
            # remove the indents
            line = re.sub(indent, '', line)

            # check if the line is an import
            if re.match(' *import .*', line) or re.match(' *from .* import .*', line):
                imports += '{0}\n'.format(line.strip())

            else:
                function_def += line

        index += 1


    # copy the import we found in the function
    script = '{0}\n'.format(imports)

    # add missing import
    if 'PicomExecutor' not in imports:
        script += '''
from torii import PicomExecutor
'''

    # add the function definition
    script += '\n{0}\n'.format(function_def)

    # add the main part
    script += '''
if __name__ == "__main__":
'''

    # we consider the first argument of the function as the executor
    # without arg, we assume the executor is instanciated by the function
    if function_args is not None and function_args.strip():
        args = function_args.split(',')
        executor = args[0]
        script += '''
    {0} =  PicomExecutor()
    '''.format(executor)

    # call the function
    script += '''
    {0}({1})
    '''.format(function_name, executor if executor else '')

    return script


class Picom(ToriiObject):

    _DICT_ENTRIES = ['phaseScripts', 'clusterScripts']

    def __init__(self, json, service, details=False):
        ToriiObject.__init__(self, json, service)


    def set_script(self, script):
        if callable(script):
            self.script = fun2script(script)
        else:
            self.script = script



from torii.data.torii_object import *
from torii.data.task import *
from torii.data.application import *
from torii.data.picom import *
from torii.data.record import *
from torii.data.cluster import *
from torii.data.struct import *
from torii.data.business_object import *

__all__ = ['Struct',
           'ToriiObject',
           'Task',
           'Application',
           'Cluster',
           'Picom',
           'Record',
           'cast_value',
           'BusinessObject']


from torii.services.service import *
from torii.services.torii_service import *
from torii.services.session_service import *
from torii.services.file_service import *
from torii.services.application_service import *
from torii.services.picom_service import *
from torii.services.task_service import *
from torii.services.graph_service import *
from torii.services.organisation_service import *
from torii.services.bo_service import *
from torii.services.profile_service import *
from torii.services.ai_service import *
from torii.services.addon_service import *

__all__ = ['Service',
           'ToriiService',
           'SessionService',
           'TaskService',
           'FileService',
           'ApplicationService',
           'PicomService',
           'ProfileService',
           'GraphService',
           'OrganisationService',
           'BusinessObjectService',
           'AiService',
           'AddonService'
           ]





from torii.exception import ToriiException
from torii.data import *
from torii.services.torii_service import ToriiService


class PicomService(ToriiService):

    def __init__(self, torii):
        """
        Create the picom service
        """
        ToriiService.__init__(self, torii=torii, name='picoms', object_class=Picom)




from torii.dao.business_object_dao import BusinessObjectDao


class DaoManager:
    """
    The manager for all data access objects (DAOs).

    All DAOs should be mapped to the 'services_ref' dictionary, where the Key is the
    first element in the REST API path and the value the uninitialized DAO class.
    """

    service_ref = {
        'bo': BusinessObjectDao,
    }

    def __init__(self, torii):
        self._torii = torii
        self._logger = self._torii.logger

        self.services = {}

        for key, bo in self.service_ref.items():
            self.services[key] = bo(torii)

    def run_service(self, intent: str, path: str, params=None, json=None, data=None, files=None, timeout=None):
        """
        Runs the intended service corresponding to the given intent and path.
        Passes the defined arguments to the service.

        :param intent: The Http method (GET, POST, PUT, DELETE or PATCH).
        :param path: The REST API path (ex.: bo/{boClass}/page).
        :param params: Optional keyword argument.
        :param json: Optional argument.
        :param data: Optional argument.
        :param files: Optional argument.
        :param timeout: Optional argument.
        :returns: The class method if exists, otherwise None.
        """

        if not self._torii.properties['use_python_dao']:
            return None

        path = path.split('/')
        service = path.pop(0)

        if service not in self.services:
            return None

        args = [intent, '/'.join(path)]
        if json: args.append(json)
        if data: args.append(data)
        if files: args.append(files)
        if timeout: args.append(timeout)
        kwargs = params or {}

        try:
            return self.services[service].init(*args, **kwargs)
        except Exception as e:
            self._logger.exception(f'Error in DAO Manager run_service')
            raise e



import time

from pymongo import ASCENDING, UpdateOne

from torii.dao.abstract_dao import AbstractDao
from torii.decorators.route_registry import RouteRegistry
from torii.exception import ToriiException


class BusinessObjectDao(AbstractDao):

    registry = RouteRegistry()

    def __init__(self, torii):
        self.db_torii = torii.mongo_database_torii
        self.db_torii_bo = torii.mongo_database_torii_bo
        self._torii = torii
        self._logger = self._torii.logger

    def init(self, intent: str, actions: str, *args, **kwargs):
        try:
            # self._logger.info(f"Calling method {intent} {actions}")
            return self.registry.call_method(self, intent, actions, *args, **kwargs)
        except Exception as e:
            self._logger.exception(f"Error in BusinessObjectDao")
            raise e

    # ===============
    # GET METHODS
    # ===============

    @registry.register('GET', '')
    def get_classes(self, explore=False):
        return self.db_torii_bo.list_collection_names()

    @registry.register('GET', '{boClass}')
    def get_documents(self, boClass, dir='ASC', sort=None, timeout=300, limit=0,
                      start=0, filter={}, include=[], raw=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)

        result = self.find_many(collection, filter, include, dir, sort, timeout, limit, start)

        if not has_permission:
            result = self._torii.user_manager.remove_no_access_documents(result)

        return result

    @registry.register('GET', '{boClass}/{id}')
    def get_document(self, boClass, id, include=[]):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)

        result = self.find_one(collection, {'_id': id}, include)

        return self._torii.user_manager.check_team_access(result) if not has_permission else result

    @registry.register('GET', '{boClass}/page')
    def get_paged_documents(self, boClass, dir='ASC', sort=None, timeout=300, limit=0,
                            start=0, filter={}, include=[], count=True, raw=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)

        result = {'result': self.find_many(collection, {}, filter, include, dir, sort, timeout, limit, start)}

        if not has_permission:
            result['result'] = self._torii.user_manager.remove_no_access_documents(result['result'])

        if count:
            result['size'] = len(result['result']) if not has_permission \
                else self.count_documents(collection, filter, start, limit)

        return result

    @registry.register('GET', '{boClass}/count')
    def count(self, boClass, timeout=300, filter={}):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)
        if not has_permission: raise ToriiException(reason)

        return self.count_documents(collection, filter)

    @registry.register('GET', '{boClass}/search/{searchedText}')
    def get_searched_objects(self, boClass, searchedText, start=0, limit=5):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)

        result = self.find_many(collection, {'$text': {'$search': searchedText}}, start=start, limit=limit)

        if not has_permission:
            result = self._torii.user_manager.remove_no_access_documents(result)

        return result

    @registry.register('GET', '{boClass}/distinct/{property}')
    def distinct(self, boClass, property, limit=100, filter={}, stats=None):
        """ TODO: Needs tests. Unknown if works as intended. """
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass)

        query = self.prepare_filter(filter)

        cursor = collection.find(query)
        cursor.limit(limit)
        cursor.distinct(property)

        if not has_permission:
            return self._torii.user_manager.remove_no_access_documents(list(cursor))

        return list(cursor)

    # ===============
    # POST METHODS
    # ===============

    @registry.register('POST', '{boClass}')
    def create_document(self, boClass, bo=None):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, read=False, write=True)
        if not has_permission: raise ToriiException(reason)

        return self.insert_one(collection, self.to_torii_object(bo))

    @registry.register('POST', '{boClass}/{id}')
    def update_document(self, boClass, id, bo=None, bulk=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        return self.update_and_retrieve_one(collection, {'_id': id}, bo)

    @registry.register('POST', '{boClass}/bulk')
    def create_documents(self, boClass, bos=None):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, read=False, write=True)
        if not has_permission: raise ToriiException(reason)

        if bos:
            self.insert_many(collection, [self.to_torii_object(bo) for bo in bos])

        return True

    # ===============
    # PUT METHODS
    # ===============

    @registry.register('PUT', '{boClass}/bulk')
    def update_documents(self, boClass, bos=None, retrieve=False, bulk=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        if not bos: return []

        updated_bos = []
        if retrieve:
            for bo in bos:
                updated_bos.append(self.update_and_retrieve_one(collection, {'_id': bo['_id']}, bo))
        else:
            if bulk:
                batch_size = 1000
                for i in range(0, len(bos), batch_size):
                    batch = bos[i:i+batch_size]
                    operations = [self._create_update_one_operation({'_id': bo['_id']}, bo) for bo in batch]
                    if operations:
                        collection.bulk_write(operations)
                return []
            else:
                for bo in bos:
                    self.update_one(collection, {'_id': bo['_id']}, bo)

        return updated_bos

    def _create_update_one_operation(self, query: dict, document: dict, filters=None):
        if filters is None:
            filters = {}
        query = self.prepare_query(query)
        query.update(self.prepare_filter(filters))
        document = self.prepare_update(document.copy())
        return UpdateOne(filter=query, update={'$set': document})

    @registry.register('PUT', '{boClass}/index')
    def create_index(self, boClass, fields=None, default=False, unique=False):
        """ TODO: Needs tests. Unknown if works as intended. """
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        if default:
            collection.create_index([('_id', ASCENDING), ('location', ASCENDING)], unique=True)

        if fields:
            collection.create_index([(item, ASCENDING) for item in fields], unique=unique)

        return True

    # ===============
    # DELETE METHODS
    # ===============

    @registry.register('DELETE', '{boClass}')
    def delete_conditional(self, boClass, filter={}, force=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, read=False, write=True)
        if not has_permission: raise ToriiException(reason)

        self.delete_many(collection, {}, filter)

        return True

    @registry.register('DELETE', '{boClass}/{id}')
    def delete_document(self, boClass, id, force=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, read=False, write=True)
        if not has_permission: raise ToriiException(reason)

        self.delete_one(collection, {'_id': id})

        return True

    @registry.register('DELETE', '{boClass}/clear/{dataset}')
    def delete_dataset(self, boClass, dataset):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, read=False, write=True)
        if not has_permission: raise ToriiException(reason)

        self.delete_many(collection, {'content.dataset': dataset})

        return True

    # ===============
    # PATCH METHODS
    # ===============

    @registry.register('PATCH', '{boClass}')
    def patch_documents(self, boClass, patchOperations=None, retrieve=False):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        updated_bos = []

        for action in patchOperations:
            obj_id = action['id']
            operation = action['op']
            key = action['key']
            value = action['value']

            if operation == 'replace':
                updated_bos.append(self.update_and_retrieve_one(collection, {'_id': obj_id}, {key: value}))
            else:
                raise ToriiException(f"Patch operation not supported: {operation}")

        return updated_bos if retrieve else []

    @registry.register('PATCH', '{boClass}/{id}')
    def patch_document(self, boClass, id, patchOperations=None):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        for action in patchOperations:
            operation = action['op']
            key = action['key']
            value = action['value']

            if operation == 'replace':
                self.update_one(collection, {'_id': id}, {key: value})
            else:
                raise ToriiException(f"Patch operation not supported: {operation}")

        return True

    @registry.register('PATCH', '{boClass}/bulk')
    def patch_documents_with_filter(self, boClass, patchOperations=None, filter={}):
        collection = self.get_collection(self.db_torii, self.db_torii_bo, boClass)
        has_permission, reason = self._torii.user_manager.check_user_permission(boClass, write=True)
        if not has_permission: raise ToriiException(reason)

        for action in patchOperations:
            obj_id = action['id']
            operation = action['op']
            key = action['key']
            value = action['value']

            if operation == 'replace':
                self.update_one(collection, {'_id': obj_id}, {key: value}, filter)
            else:
                raise ToriiException(f"Patch operation not supported: {operation}")

        return True



import copy
import sys

from torii.data.struct import Struct, struct_to_json
from torii.data.torii_object import ToriiObject
from torii.exception import ToriiException


class BusinessObject(ToriiObject):
    BASE_KEYS = ['_id', 'homogeneousDate', 'canEdit', 'canRemove', 'location', 'name', 'creationDate', 'modificationDate', 'description',
                 'wikiURI', 'tags', 'users', 'teams', 'projects', 'createdBy']

    def __init__(self, json={}, service=None):
        ToriiObject.__init__(self, json, service)

    @property
    def torii(self):
        return super(ToriiObject, self)

    def __getattr__(self, key):
        """
        Override 'getattr()' and dot operator
        :param key:  the attribute key
        :return:
        """

        if key == 'content':
            # we don't remove the 'content'
            return self
        else:
            return ToriiObject.__getattr__(self, key)

    def __setattr__(self, key, value):
        """
        Override 'setattr()' and dot operator
        :param key:  the attribute key
        :param value:
        :return:
        """

        if key == 'content':
            # we don't remove the 'content'
            if isinstance(value, dict):
                # legacy compatibility feature to change the wholme content of a BusinessObject
                for subkey, subvalue in value.items():
                    setattr(self, subkey, subvalue)
            else:
                raise Exception('Cannot set attribute "content"')
        else:
            ToriiObject.__setattr__(self, key, value)

    def __delattr__(self, key):
        """
        Override 'delattr()'
        :param key:  the attribute key
        :return:
        """

        if key == 'content':
            # we don't remove the 'content'
            pass
        else:
            ToriiObject.__delattr__(self, key)

    def from_json(self, json):
        try:
            ToriiObject.from_json(self, json, transient_entries=['content'])
            if 'content' in json:
                Struct.from_json(self, json['content'], transient_entries=self.__class__._TRANSIENT_ENTRIES, wipe=False)

            if self.__class__._DEEP_COPY_ORIG:
                self._orig = copy.deepcopy(json)
            else:
                self._orig = None
        except:
            raise ToriiException('Failed convert ToriiObject \'{0}\' to Json'
                                 .format(json['name'] if 'name' in json else 'unknown'))\
                .with_traceback(sys.exc_info()[2])

    def to_json(self, ref=False):
        try:
            if ref:
                return {'id': self.id, 'name': self.name}
            else:
                json = {
                    '_id':  self.id
                }
                for key in BusinessObject.BASE_KEYS:
                    if hasattr(self, key):
                        json[key] = struct_to_json(getattr(self, key))

                json['content'] = super(ToriiObject, self).to_json(transient_entries=BusinessObject.BASE_KEYS)

                if 'id' in json:
                    del json['id']

                return json
        except:
            raise ToriiException('Failed convert {0} from Json'.format(self))\
                .with_traceback(sys.exc_info()[2])


def _is_content_attr(spr, key):
    """
    Check if there the attribute must be stored in the payload
    :param key: an attribute key
    :return:
    """

    key_content_compatible = not key.startswith('_') and key != 'content'

    return key_content_compatible and not key in spr.keys()



#!/usr/bin/env python

from distutils.core import setup


setup(name='fse.torii',
      version='1.0.34',
      description='Torii',
      author='Fujitsu Systems Europe',
      packages=['torii', 'torii.services', 'torii.data', 'torii.dao', 'torii.decorators', 'torii.managers'],
      install_requires=[
            'requests-toolbelt==1.0.0',
            'numpy==1.24.2',
            'pymongo==4.3.3'
        ]
      )
